unit mhWindowHook;
(****************************************************************
   Copyright 2001 Mikael Holm
 ****************************************************************
 This component is FREEWARE. It is provided to you AS IS! Meaning
 that you use it at your own risk, you can't hold me responsible
 for any damage that the component may (or may not) cause. You 
 may use it as you wish in any kind of application (freeware,
 shareware, commercial). You are free to modify it to suit your 
 needs. If you do modify it, find a bug or have a suggestion for
 some improvment please notify me of the change, bug or 
 suggestion.
 ****************************************************************)
interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  // The message
  TmhMessageEvent = procedure (var Message: TMessage;var Handled: boolean) of object;

/////////////////////////////////////////////////////////////////////////////////////
// TmhWindowHook
//
  TmhWindowHook = class(TComponent)
  private
    { Private declarations }
    // Pointer To Old WndProc
    FOld : Pointer;
    // Pointer ToNew WndProc
    FNew : Pointer;
    FActive : boolean;
    FOnMessage : TmhMessageEvent;
    FWinControl: TWinControl;
    procedure FormMess(Var Msg : TMessage);
    procedure SetActive(Value: boolean);
    procedure Activate;
    procedure AddToList(P: Pointer);
    procedure DeleteFromList(P: TmhWindowHook);
    procedure HookWindow;
    procedure UnhookWindow;
  protected
    { Protected declarations }
    procedure Loaded; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    { Public declarations }

  published
    { Published declarations }
    {properties}
    property Active : boolean read FActive write SetActive;
    property WinControl : TWinControl read FWinControl write FWinControl default nil;
    {Events}
    property OnMessage : TmhMessageEvent read FOnMessage write FOnMessage;
  end;

implementation

//Global list of TmhWindowHook components
var
  gHooks : TList;

// Adds a component to the global list and creates the list if it doesn't exist
procedure TmhWindowHook.AddToList(P : Pointer);
begin
  if not assigned(gHooks) then
    gHooks := TList.Create;

  gHooks.Add(P)
end;

// Deletes a component from the global list and destroys the list if it becomes empty
procedure DeleteMe(P : Pointer);
var
  Index : integer;
begin
  Index := gHooks.IndexOf(P);
  if Index >= 0 then
    gHooks.Delete(Index);
  if gHooks.Count = 0 then
  begin
    gHooks.Free;
    gHooks := Nil;
  end;
end;

// Deletes a specific component from the global list
procedure TmhWindowHook.DeleteFromList(P : TmhWindowHook);
var
  Index : integer;
  Found : boolean;
begin
  if assigned(gHooks) then
  begin
    // If it is in the list
    if gHooks.IndexOf(P) >= 0 then
    begin
      Found := False;
      Index := gHooks.Count - 1;
      // Loop through all components in the list
      while Index >= 0 do
      begin
        // If another component hooks this components hook,
        // unhook it
        if TmhWindowHook(gHooks[Index]).FOld = FNew then
        begin
          TmhWindowHook(gHooks[Index]).FOld := FOld;
          Found := true;
        end;
        dec(Index);
      end;
      // If no other component hooks this components hook
      // just unhook it
      if Not Found then
      begin
        UnhookWindow;
      end;
      // Delete component from gobal list	
      DeleteMe(Self);
    end;
  end;
end;

// The new WndProc
procedure TmhWindowHook.FormMess(Var Msg : TMessage);
Var
  Handled : boolean;
begin

  Handled := False;
  // If the event is assigned, call it
  if assigned(FOnMessage) then
  begin
    FOnMessage(Msg,Handled);
  end;
  // If the event isn't handled yet call old WndProc
  if not Handled then
  begin
    // Don't forget to set Msg.Result!!!
    Msg.Result := CallWindowProc(FOld,WinControl.Handle,Msg.Msg,Msg.WParam,Msg.LParam);
  end;
end;

constructor TmhWindowHook.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FWinControl := Nil;
  FActive := False;
  // Make FNew a pointer to the new WndProc
  FNew := MakeObjectInstance(FormMess);
end;

procedure TmhWindowHook.Loaded;
begin
  inherited Loaded;
  if FActive then
  begin
    AddToList(Self);
    Activate;
  end;
end;

procedure TmhWindowHook.SetActive(Value: boolean);
begin
  if FActive <> Value then
  begin
    FActive := Value;
    if FActive then
    begin
      if not (csLoading in ComponentState) then
      begin
        AddToList(Self);
        Activate;
      end;
    end
    else
    begin
      DeleteFromList(Self);
    end;
  end;
end;

destructor TmhWindowHook.Destroy;
begin
  DeleteFromList(Self);
  inherited Destroy;
end;

// Activates the component
procedure TmhWindowHook.Activate;
begin
  if not (csDesigning in ComponentState) and (Owner is TControl) then
  begin
    // If the property WinControl is empty try Owner
    if not assigned(FWinControl) then
      FWinControl := TWinControl(Owner);
    HookWindow;
  end;
end;

// Hook the WinControl control
procedure TmhWindowHook.HookWindow;
begin
  if not (csDesigning in ComponentState) and (Owner is TControl) then
  begin
    if assigned(FWinControl) then
    begin
      FOld := Pointer(GetWindowLong(FWinControl.Handle,GWL_WNDPROC));
      SetWindowLong(FWinControl.Handle,GWL_WNDPROC,Integer(FNew));
    end;
  end;
end;

// Unhook the WinControl Control
procedure TmhWindowHook.UnhookWindow;
begin
  if not (csDesigning in ComponentState) and (Owner is TControl) then
  begin
    if assigned(WinControl) then
    begin
      if FOld <> Nil  then
        SetWindowLong(WinControl.Handle,GWL_WNDPROC,Integer(FOld));
    end
  end;
end;

end.
